from io import BytesIO
from unittest.mock import Mock

import pytest

from xz.io import IOAbstract, IOCombiner, IOProxy


def generate_mock(length):
    mock = Mock()
    mock._length = length  # pylint: disable=protected-access
    mock.__class__ = IOAbstract
    return mock


def test_io_combiner_seek():
    originals = [
        generate_mock(2),
        generate_mock(0),
        generate_mock(8),
    ]
    combiner = IOCombiner(*originals)

    assert combiner.tell() == 0
    assert combiner.seek(7) == 7
    assert combiner.tell() == 7

    for original in originals:
        assert not original.method_calls  # did not touch original


def test_io_combiner_append():
    # pylint: disable=protected-access
    combiner = IOCombiner(generate_mock(13), generate_mock(37))
    assert combiner._length == 50
    combiner._append(IOProxy(BytesIO(b"abcdefghij"), 0, 10))
    assert combiner._length == 60
    combiner.seek(54)
    assert combiner.read(4) == b"efgh"


def test_io_combiner_append_invalid():
    # pylint: disable=protected-access
    combiner = IOCombiner(generate_mock(13), generate_mock(37))
    assert combiner._length == 50
    with pytest.raises(TypeError):
        combiner._append(BytesIO(b"abcdefghij"))


def test_io_combiner_read():
    originals = [
        IOProxy(BytesIO(b"abc"), 0, 3),
        generate_mock(0),  # size 0, will be never used
        IOProxy(BytesIO(b"defghij"), 0, 7),
    ]
    combiner = IOCombiner(*originals)

    # read all
    originals[0].seek(2)
    originals[2].seek(4)
    combiner.seek(0)
    assert combiner.read() == b"abcdefghij"
    assert originals[0].tell() == 3
    assert originals[2].tell() == 7
    combiner.seek(4)
    assert combiner.read() == b"efghij"
    assert originals[0].tell() == 3
    assert originals[2].tell() == 7

    # read partial
    originals[0].seek(2)
    originals[2].seek(4)
    combiner.seek(1)
    assert combiner.read(6) == b"bcdefg"
    assert originals[0].tell() == 3
    assert originals[2].tell() == 4
    assert combiner.read(6) == b"hij"
    assert originals[0].tell() == 3
    assert originals[2].tell() == 7
    assert combiner.read(6) == b""
    assert originals[0].tell() == 3
    assert originals[2].tell() == 7
    assert combiner.read(6) == b""
    assert originals[0].tell() == 3
    assert originals[2].tell() == 7

    # with original seek
    combiner.seek(1)
    originals[0].seek(2)
    originals[2].seek(4)
    assert combiner.read(5) == b"bcdef"
    assert originals[0].tell() == 3
    assert originals[2].tell() == 3

    # never used at all
    assert not originals[1].method_calls



import doctest
import os
from pathlib import Path

import pytest

import xz


@pytest.fixture(autouse=True, scope="module")
def change_dir():
    old_dir = os.getcwd()
    os.chdir(Path(__file__).parent / "files")
    yield
    os.chdir(old_dir)


def _parse_readme():
    code_blocks = []
    current_code_block = ""
    current_code_block_line = None
    with (Path(__file__).parent.parent.parent / "README.md").open() as fin:
        for line_no, line in enumerate(fin):
            if line.startswith("```"):
                if current_code_block_line is None:
                    if "python" in line:
                        current_code_block_line = line_no + 1
                else:
                    code_blocks.append((current_code_block_line, current_code_block))
                    current_code_block = ""
                    current_code_block_line = None
            elif current_code_block_line is not None:
                current_code_block += line
    return code_blocks


_README_CODE_BLOCKS = _parse_readme()


@pytest.fixture(
    params=[
        pytest.param(code_block, id=f"line_{line_no}")
        for line_no, code_block in _README_CODE_BLOCKS
    ]
)
def readme_block_path(request, tmp_path):
    path = tmp_path / "block.txt"
    path.write_text(request.param)
    yield path
    path.unlink()


def test_readme(readme_block_path):  # pylint: disable=redefined-outer-name
    failure_count, test_count = doctest.testfile(
        readme_block_path,
        module_relative=False,
        extraglobs={"xz": xz},
    )
    assert failure_count == 0
    assert test_count



import pytest

from xz.utils import FloorDict


def expect_floor_dict(floordict, items):
    sorted_keys = sorted(items)
    assert len(floordict) == len(items)
    assert list(floordict) == sorted_keys
    assert list(floordict.keys()) == sorted_keys
    assert list(floordict.values()) == [items[key] for key in sorted_keys]
    assert list(floordict.items()) == [(key, items[key]) for key in sorted_keys]
    # pylint: disable=protected-access
    assert floordict._keys == sorted_keys
    assert floordict._dict == items


def test_floor_dict_empty():
    floordict = FloorDict()

    expect_floor_dict(floordict, {})

    with pytest.raises(KeyError):
        floordict[0]  # pylint: disable=pointless-statement
    with pytest.raises(KeyError):
        floordict[42]  # pylint: disable=pointless-statement


def test_floor_dict_normal():
    floordict = FloorDict()
    floordict[10] = "ten"
    floordict[50] = "fifty"
    with pytest.raises(TypeError):
        floordict["wrong type"] = "wrong type"

    expect_floor_dict(floordict, {10: "ten", 50: "fifty"})

    assert floordict[10] == "ten"

    assert floordict[42] == "ten"
    assert floordict[42, False] == "ten"
    assert floordict[42, True] == (10, "ten")

    assert floordict[50] == "fifty"
    assert floordict[1337] == "fifty"
    assert floordict.get(0) is None
    with pytest.raises(KeyError):
        floordict[0]  # pylint: disable=pointless-statement
    assert floordict.get(7) is None
    with pytest.raises(KeyError):
        floordict[7]  # pylint: disable=pointless-statement
    with pytest.raises(KeyError):
        floordict[-42]  # pylint: disable=pointless-statement
    with pytest.raises(TypeError):
        floordict["wrong type"]  # pylint: disable=pointless-statement


def test_floor_dict_override():
    floordict = FloorDict()
    floordict[10] = "ten"
    floordict[20] = "twenty"
    floordict[30] = "thirty"

    expect_floor_dict(floordict, {10: "ten", 20: "twenty", 30: "thirty"})

    floordict[20] = "two-ten"
    assert floordict[15] == "ten"
    assert floordict[20] == "two-ten"
    assert floordict[25] == "two-ten"
    assert floordict[50] == "thirty"

    expect_floor_dict(floordict, {10: "ten", 20: "two-ten", 30: "thirty"})


def test_floor_dict_del():
    floordict = FloorDict()
    floordict[10] = "ten"
    floordict[20] = "twenty"
    floordict[30] = "thirty"
    assert floordict[20] == "twenty"
    assert floordict[22] == "twenty"
    expect_floor_dict(floordict, {10: "ten", 20: "twenty", 30: "thirty"})

    del floordict[20]
    assert floordict[20] == "ten"
    assert floordict[22] == "ten"
    expect_floor_dict(floordict, {10: "ten", 30: "thirty"})

    with pytest.raises(KeyError):
        del floordict[20]
    with pytest.raises(KeyError):
        del floordict[40]


def test_floor_dict_pop():
    floordict = FloorDict()
    floordict[10] = "ten"
    floordict[20] = "twenty"
    floordict[30] = "thirty"
    assert floordict[25] == "twenty"
    expect_floor_dict(floordict, {10: "ten", 20: "twenty", 30: "thirty"})

    with pytest.raises(KeyError):
        floordict.pop(25)

    assert floordict.pop(20) == "twenty"
    expect_floor_dict(floordict, {10: "ten", 30: "thirty"})
    assert floordict[25] == "ten"


def test_floor_dict_values():
    floordict = FloorDict()
    expected = {}
    for i in range(50):
        floordict[i * 2] = str(i * 2)
        expected[i * 2] = str(i * 2)
        expect_floor_dict(floordict, expected)
        for j in range(100):
            value = min(i * 2, j - (j % 2))
            assert floordict[j] == str(value)
            assert floordict[j, True] == (value, str(value))



from io import SEEK_CUR

from xz.block import XZBlock
from xz.common import (
    XZError,
    parse_xz_footer,
    parse_xz_header,
    parse_xz_index,
    round_up,
)
from xz.io import IOCombiner, IOProxy


class XZStream(IOCombiner):
    def __init__(self, check):
        super().__init__()
        self.check = check

    @property
    def block_boundaries(self):
        return list(self._fileobjs)

    @classmethod
    def parse(cls, fileobj):
        """Parse one XZ stream from a fileobj.

        fileobj position should be right at the end of the stream when calling
        and will be moved right at the start of the stream
        """
        # footer
        fileobj.seek(-12, SEEK_CUR)
        footer = fileobj.read(12)
        check, backward_size = parse_xz_footer(footer)
        stream = cls(check)

        # index
        block_start = fileobj.seek(-12 - backward_size, SEEK_CUR)
        index = fileobj.read(backward_size)
        records = parse_xz_index(index)
        blocks_len = sum(round_up(unpadded_size) for unpadded_size, _ in records)
        block_start -= blocks_len
        for unpadded_size, uncompressed_size in records:
            block_end = block_start + round_up(unpadded_size)
            stream._append(
                XZBlock(
                    IOProxy(fileobj, block_start, block_end),
                    stream.check,
                    unpadded_size,
                    uncompressed_size,
                )
            )
            block_start = block_end

        # header
        fileobj.seek(-12 - blocks_len - backward_size, SEEK_CUR)
        header = fileobj.read(12)
        header_check = parse_xz_header(header)
        if header_check != stream.check:
            raise XZError("stream: inconsistent check value")

        fileobj.seek(-12, SEEK_CUR)
        return stream



from io import BytesIO
from unittest.mock import Mock

from xz.io import IOProxy


def test_io_proxy_seek():
    original = Mock()
    proxy = IOProxy(original, 4, 14)

    assert proxy.tell() == 0
    assert proxy.seek(7) == 7
    assert proxy.tell() == 7

    assert not original.method_calls  # did not touch original


def test_io_proxy_read():
    original = BytesIO(b"xxxxabcdefghijyyyyy")
    proxy = IOProxy(original, 4, 14)

    # read all
    original.seek(2)
    proxy.seek(0)
    assert proxy.read() == b"abcdefghij"
    assert original.tell() == 14
    proxy.seek(4)
    assert proxy.read() == b"efghij"
    assert original.tell() == 14

    # read partial
    original.seek(2)
    proxy.seek(6)
    assert proxy.read(3) == b"ghi"
    assert original.tell() == 13
    assert proxy.read(3) == b"j"
    assert original.tell() == 14
    assert proxy.read(3) == b""
    assert original.tell() == 14
    assert proxy.read(3) == b""
    assert original.tell() == 14

    # with original seek
    original.seek(2)
    proxy.seek(4)
    original.seek(1)
    assert proxy.read() == b"efghij"
    assert original.tell() == 14



from itertools import chain, product
from pathlib import Path

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--generate-integration-files",
        action="store_true",
        default=False,
        help="Test the generation of the integration files",
    )


def pytest_collection_modifyitems(config, items):
    root = Path(__file__).parent.parent
    for item in items:
        relative = Path(item.fspath).parent.relative_to(root)
        mark = relative.name
        item.add_marker(getattr(pytest.mark, mark))
    if not config.getoption("--generate-integration-files"):
        skip_mark = pytest.mark.skip(
            reason="need --generate-integration-files option to run"
        )
        for item in items:
            if "generate_integration_files" in item.keywords:
                item.add_marker(skip_mark)


# any 3 consecutive bytes is unique in _DATA_PATTERN
_DATA_PATTERN = bytes(
    chain(
        *product(
            range(65, 91),  # uppercase
            range(97, 123),  # lowercase
            range(48, 58),  # digit
        )
    )
)


@pytest.fixture(scope="session")
def data_pattern():
    return _DATA_PATTERN


@pytest.fixture(scope="session")
def data_pattern_locate():
    def locate(data):
        if len(data) < 3:
            raise ValueError("data to short")
        return (_DATA_PATTERN.index(data), len(data))

    yield locate



from io import BytesIO

import pytest

from xz.open import xz_open

# a stream with two blocks (lengths: 10, 3)
STREAM_BYTES = bytes.fromhex(
    "fd377a585a000004e6d6b4460200210116000000742fe5a3010009e299a52075"
    "74663820e2000000404506004bafe33d0200210116000000742fe5a301000299"
    "a50a0000c6687a2b8dbda0cf0002220a1b0300001b1c3777b1c467fb02000000"
    "0004595a"
)


def test_open_rb():
    fileobj = BytesIO(STREAM_BYTES)

    with xz_open(fileobj, "rb") as xz_file:
        assert xz_file._length == 13  # pylint: disable=protected-access
        assert xz_file.stream_boundaries == [0]
        assert xz_file.block_boundaries == [0, 10]

        assert xz_file.read() == b"\xe2\x99\xa5 utf8 \xe2\x99\xa5\n"

        assert xz_file.seek(9) == 9
        assert xz_file.read() == b"\xe2\x99\xa5\n"


def test_open_rt():
    fileobj = BytesIO(STREAM_BYTES)

    with xz_open(fileobj, "rt") as xz_file:
        assert xz_file.stream_boundaries == [0]
        assert xz_file.block_boundaries == [0, 10]

        assert xz_file.read() == "♥ utf8 ♥\n"

        assert xz_file.seek(9) == 9
        assert xz_file.read() == "♥\n"


def test_open_rt_file(tmp_path):
    file_path = tmp_path / "file.xz"
    file_path.write_bytes(STREAM_BYTES)

    with file_path.open("rb") as fin:
        with xz_open(fin, "rt") as xz_file:
            assert xz_file.stream_boundaries == [0]
            assert xz_file.block_boundaries == [0, 10]
            assert xz_file.fileno() == fin.fileno()

            assert xz_file.read() == "♥ utf8 ♥\n"

            assert xz_file.seek(9) == 9
            assert xz_file.read() == "♥\n"


@pytest.mark.parametrize("mode", ("rtb", "rbt"))
def test_open_mode_invalid(mode):
    fileobj = BytesIO(STREAM_BYTES)

    with pytest.raises(ValueError) as exc_info:
        xz_open(fileobj, mode)
    assert str(exc_info.value) == f"Invalid mode: {mode}"


@pytest.mark.parametrize(
    "encoding, expected",
    (
        pytest.param("utf8", "еñϲоԺε", id="utf8"),
        pytest.param("latin1", "ÐµÃ±Ï²Ð¾ÔºÎµ", id="latin1"),
    ),
)
def test_open_encoding(encoding, expected):
    fileobj = BytesIO(
        bytes.fromhex(
            "fd377a585a000000ff12d9410200210116000000742fe5a301000bd0b5c3b1cf"
            "b2d0bed4baceb50000011c0c5da447cf06729e7a010000000000595a"
        )
    )
    with xz_open(fileobj, "rt", encoding=encoding) as xz_file:
        assert xz_file.read() == expected


@pytest.mark.parametrize(
    "errors, expected",
    (
        pytest.param(None, None, id="None"),
        pytest.param("strict", None, id="strict"),
        pytest.param("ignore", "encoding", id="ignore"),
        pytest.param("replace", "en�co�di�ng", id="replace"),
        pytest.param(
            "backslashreplace", r"en\x99co\x98di\x97ng", id="backslashreplace"
        ),
    ),
)
def test_open_encoding_errors(errors, expected):
    fileobj = BytesIO(
        bytes.fromhex(
            "fd377a585a000000ff12d9410200210116000000742fe5a301000a656e99636f"
            "986469976e67000000011b0b39a7621e06729e7a010000000000595a"
        )
    )

    with xz_open(fileobj, "rt", errors=errors) as xz_file:
        if expected is None:
            with pytest.raises(ValueError):
                xz_file.read()
        else:
            assert xz_file.read() == expected


@pytest.mark.parametrize(
    "newline, expected",
    (
        pytest.param(None, ["a\n", "b\n", "c\n", "d"], id="None"),
        pytest.param("", ["a\n", "b\r", "c\r\n", "d"], id="''"),
        pytest.param("\n", ["a\n", "b\rc\r\n", "d"], id="'\n'"),
        pytest.param("\r", ["a\nb\r", "c\r", "\nd"], id="'\r'"),
        pytest.param("\r\n", ["a\nb\rc\r\n", "d"], id="'\r\n'"),
    ),
)
def test_open_newline(newline, expected):
    fileobj = BytesIO(
        bytes.fromhex(
            "fd377a585a000000ff12d9410200210116000000742fe5a3010007610a620d63"
            "0d0a64000001180840a546ac06729e7a010000000000595a"
        )
    )

    with xz_open(fileobj, "rt", newline=newline) as xz_file:
        assert xz_file.readlines() == expected


def test_open_binary_encoding():
    fileobj = BytesIO(STREAM_BYTES)
    with pytest.raises(ValueError):
        xz_open(fileobj, "rb", encoding="latin1")


def test_open_binary_encoding_errors():
    fileobj = BytesIO(STREAM_BYTES)
    with pytest.raises(ValueError):
        xz_open(fileobj, "rb", errors="ignore")


def test_open_binary_newline():
    fileobj = BytesIO(STREAM_BYTES)
    with pytest.raises(ValueError):
        xz_open(fileobj, "rb", newline="\n")



from io import SEEK_SET, BytesIO
from unittest.mock import Mock, call

import pytest

from xz.block import XZBlock
from xz.common import XZError
from xz.io import IOProxy

BLOCK_BYTES = bytes.fromhex(
    "0200210116000000742fe5a3e0006300415d00209842100431d01ab285328305"
    "7ddb5924a128599cc9911a7fcff8d59c1f6f887bcee97b1f83f1808f005de273"
    "e1a6e99a7eac4f8f632b7e43bbf1da311dce5c0000000000e7c35efa"
)


@pytest.fixture
def fileobj():
    proxy = IOProxy(BytesIO(BLOCK_BYTES), 0, 92)
    mock = Mock(wraps=proxy)
    mock.__class__ = IOProxy
    mock._length = proxy._length  # pylint: disable=protected-access
    yield mock


# pylint: disable=redefined-outer-name


def test_read_all(fileobj, data_pattern_locate):
    block = XZBlock(fileobj, 1, 89, 100)
    block.compressed_read_size = 17
    assert block.tell() == 0
    assert data_pattern_locate(block.read()) == (0, 100)

    assert fileobj.method_calls == [
        call.seek(0, SEEK_SET),
        call.read(5),  # xz padding is 12 bytes
        call.seek(5, SEEK_SET),
        call.read(17),
        call.seek(22, SEEK_SET),
        call.read(17),
        call.seek(39, SEEK_SET),
        call.read(17),
        call.seek(56, SEEK_SET),
        call.read(17),
        call.seek(73, SEEK_SET),
        call.read(17),
        # below is not needed to get the data
        # but needed to perform various checks
        # see other tests
        call.seek(90, SEEK_SET),
        call.read(17),
    ]


def test_seek_forward(fileobj, data_pattern_locate):
    block = XZBlock(fileobj, 1, 89, 100)
    block.compressed_read_size = 17
    assert block.tell() == 0

    block.seek(0)
    assert block.tell() == 0
    assert not fileobj.method_calls  # no file access
    assert data_pattern_locate(block.read(4)) == (0, 4)
    assert block.tell() == 4
    assert fileobj.method_calls == [
        call.seek(0, SEEK_SET),
        call.read(5),  # xz padding is 12 bytes
        call.seek(5, SEEK_SET),
        call.read(17),
        call.seek(22, SEEK_SET),
        call.read(17),
    ]
    fileobj.method_calls.clear()

    block.seek(10)
    assert block.tell() == 10
    assert not fileobj.method_calls  # no file access
    assert data_pattern_locate(block.read(4)) == (10, 4)
    assert block.tell() == 14
    assert not fileobj.method_calls  # no file access

    block.seek(30)
    assert block.tell() == 30
    assert fileobj.method_calls == [
        call.seek(39, SEEK_SET),
        call.read(17),
    ]
    fileobj.method_calls.clear()
    assert data_pattern_locate(block.read(4)) == (30, 4)
    assert block.tell() == 34
    assert not fileobj.method_calls  # no file access

    block.seek(60)
    assert block.tell() == 60
    assert fileobj.method_calls == [
        call.seek(56, SEEK_SET),
        call.read(17),
    ]
    fileobj.method_calls.clear()
    assert data_pattern_locate(block.read(4)) == (60, 4)
    assert block.tell() == 64
    assert not fileobj.method_calls  # no file access


def test_seek_backward(fileobj, data_pattern_locate):
    block = XZBlock(fileobj, 1, 89, 100)
    block.compressed_read_size = 17
    assert block.tell() == 0

    block.seek(60)
    assert block.tell() == 60
    assert fileobj.method_calls == [
        call.seek(0, SEEK_SET),
        call.read(5),  # xz padding is 12 bytes
        call.seek(5, SEEK_SET),
        call.read(17),
        call.seek(22, SEEK_SET),
        call.read(17),
        call.seek(39, SEEK_SET),
        call.read(17),
        call.seek(56, SEEK_SET),
        call.read(17),
    ]
    fileobj.method_calls.clear()
    fileobj.method_calls.clear()

    block.seek(40)
    assert block.tell() == 40
    assert fileobj.method_calls == [
        call.seek(0, SEEK_SET),
        call.read(5),  # xz padding is 12 bytes
        call.seek(5, SEEK_SET),
        call.read(17),
        call.seek(22, SEEK_SET),
        call.read(17),
        call.seek(39, SEEK_SET),
        call.read(17),
    ]
    fileobj.method_calls.clear()
    assert data_pattern_locate(block.read(4)) == (40, 4)
    assert block.tell() == 44
    assert not fileobj.method_calls  # no file access

    block.seek(20)
    assert block.tell() == 20
    assert fileobj.method_calls == [
        call.seek(0, SEEK_SET),
        call.read(5),  # xz padding is 12 bytes
        call.seek(5, SEEK_SET),
        call.read(17),
        call.seek(22, SEEK_SET),
        call.read(17),
    ]
    fileobj.method_calls.clear()
    assert data_pattern_locate(block.read(4)) == (20, 4)
    assert block.tell() == 24
    assert fileobj.method_calls == [
        call.seek(39, SEEK_SET),
        call.read(17),
    ]


def test_wrong_uncompressed_size_too_small(fileobj, data_pattern_locate):
    block = XZBlock(fileobj, 1, 89, 99)
    block.compressed_read_size = 17

    # read all but last byte
    assert data_pattern_locate(block.read(98)) == (0, 98)

    # read last byte
    with pytest.raises(XZError) as exc_info:
        block.read(1)
    assert str(exc_info.value) == "block: error while decompressing: Corrupt input data"


def test_wrong_uncompressed_size_too_big(fileobj, data_pattern_locate):
    block = XZBlock(fileobj, 1, 89, 101)
    block.compressed_read_size = 17

    # read all but last byte
    assert data_pattern_locate(block.read(100)) == (0, 100)

    # read last byte
    print("-" * 10)
    with pytest.raises(XZError) as exc_info:
        block.read(1)
    assert str(exc_info.value) == "block: error while decompressing: Corrupt input data"


def test_wrong_block_padding(data_pattern_locate):
    fileobj = IOProxy(BytesIO(BLOCK_BYTES[:-5] + b"\xff" + BLOCK_BYTES[-4:]), 0, 92)
    block = XZBlock(fileobj, 1, 89, 100)
    block.compressed_read_size = 17

    # read all but last byte
    assert data_pattern_locate(block.read(99)) == (0, 99)

    # read last byte
    with pytest.raises(XZError) as exc_info:
        block.read(1)
    assert str(exc_info.value) == "block: error while decompressing: Corrupt input data"


def test_wrong_check(data_pattern_locate):
    fileobj = IOProxy(BytesIO(BLOCK_BYTES[:-4] + b"\xff" * 4), 0, 92)

    block = XZBlock(fileobj, 1, 89, 100)
    block.compressed_read_size = 17

    # read all but last byte
    assert data_pattern_locate(block.read(99)) == (0, 99)

    # read last byte
    with pytest.raises(XZError) as exc_info:
        block.read(1)
    assert str(exc_info.value) == "block: error while decompressing: Corrupt input data"


def test_truncated_data(fileobj):
    block = XZBlock(fileobj, 1, 89, 100)
    block.compressed_read_size = 17
    block.compressed_fileobj = IOProxy(
        BytesIO(
            bytes.fromhex(
                # header
                "fd377a585a0000016922de36"
                # one block (truncated)
                "0200210116000000742fe5a301000941"
            )
        ),
        0,
        28,
    )

    with pytest.raises(XZError) as exc_info:
        block.read()
    assert str(exc_info.value) == "block: data eof"


def test_decompressor_eof(data_pattern_locate):
    fileobj = IOProxy(
        BytesIO(
            bytes.fromhex(
                # one block
                "0200210116000000742fe5a301000941"
                "6130416131416132410000004e4aa467"
                # index
                "00011e0aea6312149042990d0100"
                # stream footer
                "00000001595a"
            )
        ),
        0,
        52,
    )

    # real uncompressed size is 10, not 11
    # it is changed to trigger the error case we are testing here
    block = XZBlock(fileobj, 1, 30, 11)
    block.compressed_read_size = 17

    # read all but last byte
    assert data_pattern_locate(block.read(10)) == (0, 10)

    # read last byte
    with pytest.raises(XZError) as exc_info:
        block.read(1)
    assert str(exc_info.value) == "block: decompressor eof"



from binascii import crc32 as crc32int
from struct import pack, unpack

HEADER_MAGIC = b"\xfd7zXZ\x00"
FOOTER_MAGIC = b"YZ"


class XZError(Exception):
    pass


def encode_mbi(value):
    data = bytearray()
    while value >= 0x80:
        data.append((value | 0x80) & 0xFF)
        value >>= 7
    data.append(value)
    return data


def decode_mbi(data):
    value = 0
    for size, byte in enumerate(data):
        value |= (byte & 0x7F) << (size * 7)
        if not byte & 0x80:
            return (size + 1, value)
    raise XZError("invalid mbi")


def crc32(data):
    return pack("<I", crc32int(data))


def round_up(value):
    remainder = value % 4
    if remainder:
        return value - remainder + 4
    return value


def pad(value):
    return b"\x00" * (round_up(value) - value)


def create_xz_header(check):
    if not 0 <= check <= 0xF:
        raise XZError("header check")
    # stream header
    flags = pack("<BB", 0, check)
    return HEADER_MAGIC + flags + crc32(flags)


def create_xz_index_footer(check, records):
    if not 0 <= check <= 0xF:
        raise XZError("footer check")
    # index
    index = b"\x00"
    index += encode_mbi(len(records))
    for unpadded_size, uncompressed_size in records:
        index += encode_mbi(unpadded_size)
        index += encode_mbi(uncompressed_size)
    index += pad(len(index))
    index += crc32(index)
    # stream footer
    footer = pack("<IBB", (len(index) // 4) - 1, 0, check)
    footer = crc32(footer) + footer + FOOTER_MAGIC
    return index + footer


def parse_xz_header(header):
    if len(header) != 12:
        raise XZError("header length")
    if header[:6] != HEADER_MAGIC:
        raise XZError("header magic")
    if crc32(header[6:8]) != header[8:12]:
        raise XZError("header crc32")
    flag_first_byte, check = unpack("<BB", header[6:8])
    if flag_first_byte or not 0 <= check <= 0xF:
        raise XZError("header flags")
    return check


def parse_xz_index(index):
    if len(index) < 8 or len(index) % 4:
        raise XZError("index length")
    index = memoryview(index)
    if index[0]:
        raise XZError("index indicator")
    if crc32(index[:-4]) != index[-4:]:
        raise XZError("index crc32")
    size, nb_records = decode_mbi(index[1:])
    index = index[1 + size : -4]
    # records
    records = []
    for _ in range(nb_records):
        if not index:
            raise XZError("index size")
        size, unpadded_size = decode_mbi(index)
        if not unpadded_size:
            raise XZError("index record unpadded size")
        index = index[size:]
        if not index:
            raise XZError("index size")
        size, uncompressed_size = decode_mbi(index)
        if not uncompressed_size:
            raise XZError("index record uncompressed size")
        index = index[size:]
        records.append((unpadded_size, uncompressed_size))
    # index padding
    if any(index):
        raise XZError("index padding")
    return records


def parse_xz_footer(footer):
    if len(footer) != 12:
        raise XZError("footer length")
    if footer[10:12] != FOOTER_MAGIC:
        raise XZError("footer magic")
    if crc32(footer[4:10]) != footer[:4]:
        raise XZError("footer crc32")
    backward_size, flag_first_byte, check = unpack("<IBB", footer[4:10])
    backward_size = (backward_size + 1) * 4
    if flag_first_byte or not 0 <= check <= 0xF:
        raise XZError("footer flags")
    return (check, backward_size)



from io import SEEK_CUR, SEEK_END
import os

from xz.common import XZError
from xz.io import IOCombiner
from xz.stream import XZStream


class XZFile(IOCombiner):
    """A file object providing transparent XZ (de)compression.

    An XZFile can act as a wrapper for an existing file object, or
    refer directly to a named file on disk.

    Note that XZFile provides a *binary* file interface - data read
    is returned as bytes, and data to be written must be given as bytes.
    Use xz.open if you want a *text* file interface.
    """

    def __init__(self, filename=None, mode="r"):
        """Open an XZ file in binary mode.

        filename can be either an actual file name (given as a str,
        bytes, or PathLike object), in which case the named file is
        opened, or it can be an existing file object to read from or
        write to.

        mode can be "r" for reading (default).
        It is equivalent to "rb".
        """

        self.close_fileobj = False
        self.mode = mode
        if self.mode.endswith("b"):
            self.mode = self.mode[:-1]

        super().__init__()

        if isinstance(filename, (str, bytes, os.PathLike)):
            # pylint: disable=consider-using-with, unspecified-encoding
            self.fileobj = open(filename, self.mode + "b")
            self.close_fileobj = True
        elif hasattr(filename, "read"):
            self.fileobj = filename
        else:
            raise TypeError("filename must be a str, bytes, file or PathLike object")

        # we only support read for now
        if self.mode in ("r", "rb"):
            self._init_parse()
        else:
            raise ValueError(f"invalid mode: {mode}")

    def fileno(self):
        try:
            return self.fileobj.fileno()
        except AttributeError as ex:
            raise OSError("underlying file object has no fileno") from ex

    def close(self):
        try:
            super().close()
        finally:
            if self.close_fileobj:
                self.fileobj.close()

    @property
    def stream_boundaries(self):
        return list(self._fileobjs)

    @property
    def block_boundaries(self):
        return [
            stream_pos + block_boundary
            for stream_pos, stream in self._fileobjs.items()
            for block_boundary in stream.block_boundaries
        ]

    def _init_parse(self):
        self.fileobj.seek(0, SEEK_END)

        streams = []

        while self.fileobj.tell():
            if self.fileobj.tell() % 4:
                raise XZError("file: invalid size")
            self.fileobj.seek(-4, SEEK_CUR)
            if any(self.fileobj.read(4)):
                streams.append(XZStream.parse(self.fileobj))
            else:
                self.fileobj.seek(-4, SEEK_CUR)  # stream padding

        if not streams:
            raise XZError("file: no streams")

        while streams:
            self._append(streams.pop())



from xz.io import IOStatic


def test_io_proxy_read():
    static = IOStatic(b"abcdefghij")

    # read all
    static.seek(0)
    assert static.read() == b"abcdefghij"
    static.seek(4)
    assert static.read() == b"efghij"

    # read partial
    static.seek(6)
    assert static.read(3) == b"ghi"
    assert static.read(3) == b"j"
    assert static.read(3) == b""
    assert static.read(3) == b""



from io import DEFAULT_BUFFER_SIZE, SEEK_SET
from lzma import FORMAT_XZ, LZMADecompressor, LZMAError

from xz.common import XZError, create_xz_header, create_xz_index_footer
from xz.io import IOAbstract, IOCombiner, IOStatic


class XZBlock(IOAbstract):
    compressed_read_size = DEFAULT_BUFFER_SIZE

    def __init__(self, fileobj, check, unpadded_size, uncompressed_size):
        super().__init__(uncompressed_size)
        self.compressed_fileobj = IOCombiner(
            IOStatic(create_xz_header(check)),
            fileobj,
            IOStatic(
                create_xz_index_footer(check, [(unpadded_size, uncompressed_size)])
            ),
        )
        self._decompressor_reset()

    def _decompressor_reset(self):
        self.compressed_fileobj.seek(0, SEEK_SET)
        self.decompressor = LZMADecompressor(format=FORMAT_XZ)

    def _decompressor_read(self, size):
        # pylint: disable=using-constant-test
        if self.decompressor.eof:
            raise XZError("block: decompressor eof")
        if self.decompressor.needs_input:
            data_input = self.compressed_fileobj.read(self.compressed_read_size)
            if not data_input:
                raise XZError("block: data eof")
        else:
            data_input = b""
        return self.decompressor.decompress(data_input, size)

    def seek(self, *args):
        old_pos = self._pos
        super().seek(*args)
        pos_diff = self._pos - old_pos
        if pos_diff < 0:
            self._decompressor_reset()
            old_pos = 0
            pos_diff = self._pos
        if pos_diff > 0:
            self._pos = old_pos
            self.read(pos_diff)

    def _read(self, size):
        try:
            data_output = self._decompressor_read(size)

            if self._pos + len(data_output) == self._length:
                # we reached the end of the block
                # according to the XZ specification, we must check the
                # remaining bytes of the block; this is mainly performed by the
                # decompressor itself when we consume it
                while not self.decompressor.eof:
                    if self._decompressor_read(1):
                        raise LZMAError("Corrupt input data")

            return data_output

        except LZMAError as ex:
            raise XZError(f"block: error while decompressing: {ex}") from ex



from io import SEEK_CUR, SEEK_END, SEEK_SET, IOBase

from xz.utils import FloorDict


class IOAbstract(IOBase):
    def __init__(self, length):
        super().__init__()
        self._pos = 0
        self._length = length

    def __repr__(self):
        return f"<{self.__class__.__name__} object at {hex(hash(self))}>"

    def seek(self, pos, whence=SEEK_SET):
        """Change stream position.

        Change the stream position to byte offset pos. Argument pos is
        interpreted relative to the position indicated by whence. Values
        for whence are ints:

        * 0 -- start of stream (the default); offset should be zero or positive
        * 1 -- current stream position; offset may be negative
        * 2 -- end of stream; offset should be negative

        Return an int indicating the new absolute position.
        """
        if whence == SEEK_SET:
            pass
        elif whence == SEEK_CUR:
            pos += self._pos
        elif whence == SEEK_END:
            pos += self._length
        else:
            raise ValueError("unsupported whence value")
        if 0 <= pos <= self._length:
            self._pos = pos
            return self._pos
        raise ValueError("invalid seek position")

    def tell(self):
        """Return an int indicating the current stream position."""
        return self._pos

    def readable(self):
        """Return a bool indicating whether object was opened for reading."""
        return True

    def seekable(self):
        """Return a bool indicating whether object supports random access."""
        return True

    def writable(self):
        """Return a bool indicating whether object was opened for writing."""
        return False

    def read(self, size=-1):
        """Read at most size bytes, returned as a bytes object.

        If the size argument is negative, read until EOF is reached.
        Return an empty bytes object at EOF.
        """
        if size < 0:
            size = self._length
        size = min(size, self._length - self._pos)
        parts = []
        while size:
            data = self._read(size)  # do not stop if data is empty
            parts.append(data)
            size -= len(data)
            self._pos += len(data)
        return b"".join(parts)

    def _read(self, size):  # pragma: no cover
        """Read and return up to size bytes, where size is an int.

        The size will not exceed the number of bytes between self._pos and
        self_length. This should prevent to deal with EOF.

        This function can return less bytes than size, in which case it will be
        called again. This includes being able to return an empty bytes object.
        """
        raise NotImplementedError  # must be overridden


class IOStatic(IOAbstract):
    def __init__(self, data):
        self.data = bytearray(data)
        super().__init__(len(self.data))

    def _read(self, size):
        return self.data[self._pos : self._pos + size]


class IOProxy(IOAbstract):
    def __init__(self, fileobj, start, end):
        super().__init__(end - start)
        self.fileobj = fileobj
        self.start = start

    def _read(self, size):
        self.fileobj.seek(self.start + self._pos, SEEK_SET)
        return self.fileobj.read(size)  # size already restricted by caller


class IOCombiner(IOAbstract):
    def __init__(self, *fileobjs):
        super().__init__(0)
        self._fileobjs = FloorDict()
        for fileobj in fileobjs:
            self._append(fileobj)

    def _read(self, size):
        start, fileobj = self._fileobjs[self._pos, True]
        fileobj.seek(self._pos - start, SEEK_SET)
        return fileobj.read(size)

    def _append(self, fileobj):
        if not isinstance(fileobj, IOAbstract):
            raise TypeError
        self._fileobjs[self._length] = fileobj  # override empty streams
        self._length += fileobj._length  # pylint: disable=protected-access



from hashlib import sha1
import subprocess

import pytest


@pytest.mark.generate_integration_files
def test(integration_case, data_pattern):
    xz_path, metadata = integration_case

    expected_hash = sha1(xz_path.read_bytes())

    # note that we override current xz file
    # this allows to create new integration files from json metadata
    data = memoryview(data_pattern)
    with xz_path.open("wb") as fout:
        for step in metadata["generate"]:
            step_data_len = step.get("length", 0)
            step_data = data[:step_data_len]
            data = data[step_data_len:]
            fout.write(
                subprocess.run(
                    step["cmd"].split(" "),
                    input=step_data,
                    stdout=subprocess.PIPE,
                    check=True,
                ).stdout
            )
    assert not data

    generated_hash = sha1(xz_path.read_bytes())

    assert generated_hash.hexdigest() == expected_hash.hexdigest()



from lzma import CHECK_CRC32, CHECK_CRC64, CHECK_NONE, CHECK_SHA256

import pytest

from xz.common import (
    XZError,
    create_xz_header,
    create_xz_index_footer,
    decode_mbi,
    encode_mbi,
    pad,
    parse_xz_footer,
    parse_xz_header,
    parse_xz_index,
    round_up,
)

MBI_CASE = tuple(
    pytest.param(value, data, id=hex(value))
    for value, data in (
        (0, "00"),
        (1, "01"),
        ((1 << 7) - 1, "7f"),
        (1 << 7, "8001"),
        ((1 << 7 * 2) - 1, "ff7f"),
        (1 << 7 * 2, "808001"),
        ((1 << 7 * 3) - 1, "ffff7f"),
        (1 << 7 * 3, "80808001"),
        ((1 << 7 * 10) - 1, "ffffffffffffffffff7f"),
        (1 << 7 * 10, "8080808080808080808001"),
        (9, "09"),
        (99, "63"),
        (999, "e707"),
        (9999, "8f4e"),
        (99999, "9f8d06"),
        (999999, "bf843d"),
        (9999999, "fface204"),
        (99999999, "ffc1d72f"),
        (999999999, "ff93ebdc03"),
    )
)


@pytest.mark.parametrize("value, data", MBI_CASE)
def test_encode_mbi(value, data):
    assert encode_mbi(value) == bytes.fromhex(data)


@pytest.mark.parametrize("value, data", MBI_CASE)
def test_decode_mbi(value, data):
    assert decode_mbi(bytes.fromhex(data) + b"\xff\x00" * 10) == (len(data) // 2, value)


@pytest.mark.parametrize("data", ("", "81828384"), ids=("empty", "truncated"))
def test_decode_mbi_invalid(data):
    with pytest.raises(XZError) as exc_info:
        decode_mbi(bytes.fromhex(data))
    assert str(exc_info.value) == "invalid mbi"


@pytest.mark.parametrize(
    "value, expected",
    ((0, 0), (1, 4), (2, 4), (3, 4), (4, 4), (5, 8), (6, 8), (7, 8), (8, 8)),
)
def test_round_up(value, expected):
    assert round_up(value) == expected


@pytest.mark.parametrize(
    "value, padding",
    (
        (0, ""),
        (1, "000000"),
        (2, "0000"),
        (3, "00"),
        (4, ""),
        (5, "000000"),
        (6, "0000"),
        (7, "00"),
        (8, ""),
    ),
)
def test_pad(value, padding):
    assert pad(value) == bytes.fromhex(padding)
    data = b"B" * value
    data += bytes.fromhex(padding)
    assert not len(data) % 4


XZ_HEADER_CASES = (
    pytest.param(CHECK_NONE, "fd377a585a000000ff12d941", id="check_none"),
    pytest.param(CHECK_CRC32, "fd377a585a0000016922de36", id="check_crc32"),
    pytest.param(CHECK_CRC64, "fd377a585a000004e6d6b446", id="check_crc64"),
    pytest.param(CHECK_SHA256, "fd377a585a00000ae1fb0ca1", id="check_sha256"),
)


@pytest.mark.parametrize("check, data", XZ_HEADER_CASES)
def test_create_xz_header(check, data):
    assert create_xz_header(check) == bytes.fromhex(data)


def test_create_xz_header_invalid_check():
    with pytest.raises(XZError) as exc_info:
        create_xz_header(17)
    assert str(exc_info.value) == "header check"


@pytest.mark.parametrize("check, data", XZ_HEADER_CASES)
def test_parse_xz_header(check, data):
    assert parse_xz_header(bytes.fromhex(data)) == check


@pytest.mark.parametrize(
    "data, message",
    (
        ("fd377a585a0000016922de3600", "header length"),
        ("f1377a585a000000ff12d941", "header magic"),
        ("fd377a585a0000016942de36", "header crc32"),
        ("fd377a585a0000110d32692b", "header flags"),
        ("fd377a585a0001012813c52f", "header flags"),
        ("fd377a585a00100138301c7c", "header flags"),
    ),
)
def test_parse_xz_header_invalid(data, message):
    with pytest.raises(XZError) as exc_info:
        parse_xz_header(bytes.fromhex(data))
    assert str(exc_info.value) == message


XZ_INDEX_CASES = (
    # all have check=1
    pytest.param([], "000000001cdf4421", id="empty"),
    pytest.param([(24, 4)], "000118046be9f0a5", id="one-small-block"),
    pytest.param([(2062, 20280)], "00018e10b89e010039f45fb1", id="one-big-block"),
    pytest.param(
        [(73, 60), (73, 60), (73, 60), (56, 30)],
        "0004493c493c493c381e0000b6ec1657",
        id="several-small-blocks",
    ),
    pytest.param(
        [(1, 2), (11, 2222), (1111, 22222222), (11111111, 2222222222222222)],
        "000401020bae11d7088eabcc0ac795a6058ec7abf196a3f903000000c9647142",
        id="several-blocks-various-sizes",
    ),
)


@pytest.mark.parametrize("records, data", XZ_INDEX_CASES)
def test_create_xz_index(records, data):
    assert create_xz_index_footer(1, records)[:-12] == bytes.fromhex(data)


@pytest.mark.parametrize("records, data", XZ_INDEX_CASES)
def test_parse_xz_index(records, data):
    assert parse_xz_index(bytes.fromhex(data)) == records


@pytest.mark.parametrize(
    "data, message",
    (
        ("0000001cdf4421", "index length"),
        ("420000001cdf4421", "index indicator"),
        ("000000001cdf4221", "index crc32"),
        ("000218043257b6a7", "index size"),
        ("000100043271eb27", "index record unpadded size"),
        ("000188047163b1d4", "index size"),
        ("000104002f70ea44", "index record uncompressed size"),
        ("000180180400420096a658c0", "index padding"),
    ),
)
def test_parse_xz_index_invalid(data, message):
    with pytest.raises(XZError) as exc_info:
        parse_xz_index(bytes.fromhex(data))
    assert str(exc_info.value) == message


XZ_FOOTER_CASES = (
    # all have backward_size=8 (i.e. no blocks)
    pytest.param(CHECK_NONE, "06729e7a010000000000595a", id="check_none"),
    pytest.param(CHECK_CRC32, "9042990d010000000001595a", id="check_crc32"),
    pytest.param(CHECK_CRC64, "1fb6f37d010000000004595a", id="check_crc64"),
    pytest.param(CHECK_SHA256, "189b4b9a01000000000a595a", id="check_sha256"),
)


@pytest.mark.parametrize("check, data", XZ_FOOTER_CASES)
def test_create_xz_footer(check, data):
    assert create_xz_index_footer(check, [])[-12:] == bytes.fromhex(data)


def test_create_xz_footer_invalid_check():
    with pytest.raises(XZError) as exc_info:
        create_xz_index_footer(17, [])
    assert str(exc_info.value) == "footer check"


@pytest.mark.parametrize("check, data", XZ_FOOTER_CASES)
def test_parse_xz_footer(check, data):
    assert parse_xz_footer(bytes.fromhex(data)) == (check, 8)


@pytest.mark.parametrize(
    "data, message",
    (
        ("009042990d010000000001595a", "footer length"),
        ("9042990d0100000000015959", "footer magic"),
        ("9042090d010000000001595a", "footer crc32"),
        ("f4522e10010000000011595a", "footer flags"),
        ("d1738214010000000101595a", "footer flags"),
        ("c1505b47010000001001595a", "footer flags"),
    ),
)
def test_parse_xz_footer_invalid(data, message):
    with pytest.raises(XZError) as exc_info:
        parse_xz_footer(bytes.fromhex(data))
    assert str(exc_info.value) == message



from io import SEEK_CUR, SEEK_END, BytesIO
from unittest.mock import Mock, call

import pytest

from xz.common import XZError
from xz.stream import XZStream

# a stream with two blocks (lengths: 100, 90)
STREAM_BYTES = bytes.fromhex(
    "fd377a585a0000016922de360200210116000000742fe5a3e0006300415d0020"
    "9842100431d01ab2853283057ddb5924a128599cc9911a7fcff8d59c1f6f887b"
    "cee97b1f83f1808f005de273e1a6e99a7eac4f8f632b7e43bbf1da311dce5c00"
    "00000000e7c35efa0200210116000000742fe5a3e00059003d5d00320cc42641"
    "c8b91ac7908be7e635b8e7d681d74b683cde914399f8de5460dc672363f1e067"
    "5a3ebac9f427ecbebcb94552c0dba85b26950f0ac98b22390000000031f4ee87"
    "00025964555a0000041276283e300d8b020000000001595a"
)


def test_parse(data_pattern_locate):
    fileobj = Mock(wraps=BytesIO(b"\xff" * 1000 + STREAM_BYTES + b"\xee" * 1000))
    fileobj.seek(-1000, SEEK_END)  # move at the end of the stream
    fileobj.method_calls.clear()

    # parse stream
    stream = XZStream.parse(fileobj)
    assert stream.check == 1
    assert stream._length == 190  # pylint: disable=protected-access
    assert stream.block_boundaries == [0, 100]

    # make sure we don't read the blocks
    assert fileobj.method_calls == [
        call.seek(-12, SEEK_CUR),
        call.read(12),
        call.seek(-24, SEEK_CUR),
        call.read(12),
        call.seek(-204, SEEK_CUR),  # blocks are skipped over here
        call.read(12),
        call.seek(-12, SEEK_CUR),
    ]

    # fileobj should be at the begining of the stream
    assert fileobj.tell() == 1000

    # read from start
    assert data_pattern_locate(stream.read(20)) == (0, 20)

    # read from middle of a block
    stream.seek(40)
    assert data_pattern_locate(stream.read(20)) == (40, 20)

    # read accross two blocks
    stream.seek(90)
    assert data_pattern_locate(stream.read(20)) == (90, 20)

    # read middle of an other block
    stream.seek(160)
    assert data_pattern_locate(stream.read(20)) == (160, 20)

    # go backward and read
    stream.seek(130)
    assert data_pattern_locate(stream.read(20)) == (130, 20)

    # read in previous block (going backward from last read in that block)
    stream.seek(60)
    assert data_pattern_locate(stream.read(20)) == (60, 20)

    # read until end
    stream.seek(170)
    assert data_pattern_locate(stream.read()) == (170, 20)


def test_invalid_stream_flags_missmatch():
    fileobj = BytesIO(
        bytes.fromhex(
            "fd377a585a000004e6d6b446000000001cdf44219042990d010000000001595a"
        )
    )
    fileobj.seek(0, SEEK_END)  # move at the end of the stream
    with pytest.raises(XZError) as exc_info:
        XZStream.parse(fileobj)
    assert str(exc_info.value) == "stream: inconsistent check value"



from io import BytesIO
import os

import pytest

from xz.common import XZError
from xz.file import XZFile

FILE_BYTES = bytes.fromhex(
    # stream 1: two blocks (lengths: 100, 90)
    "fd377a585a0000016922de360200210116000000742fe5a3e0006300415d0020"
    "9842100431d01ab2853283057ddb5924a128599cc9911a7fcff8d59c1f6f887b"
    "cee97b1f83f1808f005de273e1a6e99a7eac4f8f632b7e43bbf1da311dce5c00"
    "00000000e7c35efa0200210116000000742fe5a3e00059003d5d00320cc42641"
    "c8b91ac7908be7e635b8e7d681d74b683cde914399f8de5460dc672363f1e067"
    "5a3ebac9f427ecbebcb94552c0dba85b26950f0ac98b22390000000031f4ee87"
    "00025964555a0000041276283e300d8b020000000001595a"
    # stream padding
    "0000000000000000000000000000000000000000000000000000000000000000"
    "000000000000000000000000000000000000000000000000"
    # stream 2: 4 blocks (lengths: 60, 60, 60, 30)
    "fd377a585a000004e6d6b4460200210116000000742fe5a3e0003b002d5d0033"
    "8cc42671c8b91ac7908be7e635b8e7d684446f683cde914399f8de5460dc6723"
    "63fa4300e3c2a1e6de9cc32300000000b4016504474e01bb0200210116000000"
    "742fe5a3e0003b002d5d00348cc42691c8b91ac7908be7e635b8e7d685e28768"
    "3cde914399f8de5460dc672363fa43011d1be020109cdc67000000006cbf8baa"
    "964240df0200210116000000742fe5a3e0003b002d5d00358cc426b1c8b91ac7"
    "908be7e635b8e7d687809f683cde914399f8de5460dc672363fa430154fa22f1"
    "0db14a8700000000a08ece54a2e123cc0200210116000000742fe5a3e0001d00"
    "1c5d00368cc426d1c8b91ac7908be7e635b8e7d6891eb7683cdddf0721800000"
    "194fe77b8383eba10004493c493c493c381e0000b6ec165714173b3003000000"
    "0004595a"
    # stream padding
    "0000000000000000000000000000000000000000000000000000000000000000"
    "0000000000000000000000000000000000000000000000000000000000000000"
    "000000000000000000000000"
)


@pytest.mark.parametrize("filetype", ("fileobj", "filename", "path"))
def test_read(filetype, tmp_path, data_pattern_locate):
    if filetype == "fileobj":
        filename = BytesIO(FILE_BYTES)
    else:
        filename = tmp_path / "archive.xz"
        filename.write_bytes(FILE_BYTES)
        if filetype == "filename":
            filename = os.fspath(filename)

    with XZFile(filename) as xzfile:
        assert xzfile._length == 400  # pylint: disable=protected-access
        assert xzfile.stream_boundaries == [0, 190]
        assert xzfile.block_boundaries == [0, 100, 190, 250, 310, 370]

        # read from start
        assert data_pattern_locate(xzfile.read(20)) == (0, 20)

        # read from middle of a block
        xzfile.seek(40)
        assert data_pattern_locate(xzfile.read(20)) == (40, 20)

        # read accross two blocks
        xzfile.seek(90)
        assert data_pattern_locate(xzfile.read(20)) == (90, 20)

        # read middle of an other block
        xzfile.seek(160)
        assert data_pattern_locate(xzfile.read(20)) == (160, 20)

        # read accross two streams
        xzfile.seek(180)
        assert data_pattern_locate(xzfile.read(20)) == (180, 20)

        # read middle of an other block
        xzfile.seek(320)
        assert data_pattern_locate(xzfile.read(20)) == (320, 20)

        # read accross two blocks
        xzfile.seek(360)
        assert data_pattern_locate(xzfile.read(20)) == (360, 20)

        # read until the end
        assert data_pattern_locate(xzfile.read()) == (380, 20)

        # go backward and read
        xzfile.seek(210)
        assert data_pattern_locate(xzfile.read(20)) == (210, 20)

        # read in previous stream (going backward from last read in that stream)
        xzfile.seek(60)
        assert data_pattern_locate(xzfile.read(20)) == (60, 20)

        # read all
        xzfile.seek(0)
        assert data_pattern_locate(xzfile.read()) == (0, 400)


@pytest.mark.parametrize("mode", ("r", "rb"))
def test_read_with_mode(mode, data_pattern_locate):
    filename = BytesIO(FILE_BYTES)

    with XZFile(filename, mode=mode) as xzfile:
        assert xzfile._length == 400  # pylint: disable=protected-access
        assert data_pattern_locate(xzfile.read(20)) == (0, 20)


def test_read_invalid_stream_padding():
    filename = BytesIO(FILE_BYTES + b"\x00" * 3)

    with pytest.raises(XZError) as exc_info:
        XZFile(filename)
    assert str(exc_info.value) == "file: invalid size"


def test_read_invalid_filename_type():
    with pytest.raises(TypeError) as exc_info:
        XZFile(42)
    assert (
        str(exc_info.value) == "filename must be a str, bytes, file or PathLike object"
    )


@pytest.mark.parametrize("data", (b"", b"\x00" * 100), ids=("empty", "only-padding"))
def test_read_no_stream(data):
    filename = BytesIO(data)

    with pytest.raises(XZError) as exc_info:
        XZFile(filename)
    assert str(exc_info.value) == "file: no streams"


@pytest.mark.parametrize(
    "mode",
    (
        "rt",
        "r+",
        "r+b",
        "r+t",
        "w",
        "wb",
        "wt",
        "w+",
        "w+b",
        "w+t",
        "x",
        "xb",
        "xt",
        "x+",
        "x+b",
        "x+t",
        "a",
        "ab",
        "at",
        "a+",
        "a+b",
        "a+t",
        "what-is-this",
    ),
)
def test_invalid_mode(mode):
    filename = BytesIO(FILE_BYTES)
    with pytest.raises(ValueError) as exc_info:
        XZFile(filename, mode)
    assert str(exc_info.value) == f"invalid mode: {mode}"


def test_fileno(tmp_path):
    file_path = tmp_path / "file.xz"
    file_path.write_bytes(FILE_BYTES)

    with file_path.open("rb") as fin:
        with XZFile(fin) as xzfile:
            assert xzfile.fileno() == fin.fileno()


def test_fileno_error(tmp_path):
    file_path = tmp_path / "file.xz"
    file_path.write_bytes(FILE_BYTES)

    class FakeFile:
        def __init__(self, fileobj):
            self.fileobj = fileobj

        def read(self, *args, **kwargs):
            return self.fileobj.read(*args, **kwargs)

        def seek(self, *args, **kwargs):
            return self.fileobj.seek(*args, **kwargs)

        def tell(self, *args, **kwargs):
            return self.fileobj.tell(*args, **kwargs)

    with file_path.open("rb") as fin:
        with XZFile(FakeFile(fin)) as xzfile:
            with pytest.raises(OSError) as exc_info:
                xzfile.fileno()
            assert str(exc_info.value) == "underlying file object has no fileno"



from bisect import bisect_right, insort_right
from collections.abc import MutableMapping


class FloorDict(MutableMapping):
    """A dict where keys are int, and accessing a key will use the closest lower one.

    Differences from dict:
     - keys must be int
     - obj[key] will return the value whose key is the closest one which is lower or equal to key
     - obj[key, True] will return the real index of the value in the form (index, obj[key])
    """

    def __init__(self):
        self._dict = {}
        self._keys = []  # sorted

    def __repr__(self):
        return f"FloorDict<{self._dict!r}>"

    def __iter__(self):
        return iter(self._keys)

    def __len__(self):
        return len(self._keys)

    def _key_index(self, key):
        index = bisect_right(self._keys, key) - 1
        if index < 0:
            raise KeyError(key)
        return index

    def __getitem__(self, key):
        with_index = False
        if isinstance(key, tuple) and len(key) == 2:
            key, with_index = key
        if not isinstance(key, int):
            raise TypeError("Invalid key")
        index = self._keys[self._key_index(key)]
        value = self._dict[index]
        if with_index:
            return (index, value)
        return value

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise TypeError("Invalid key")
        if key not in self._dict:  # prevent duplicates in _keys
            insort_right(self._keys, key)
        self._dict[key] = value

    def __delitem__(self, key):
        del self._dict[key]
        # the key is an exact index (otherwise KeyError raised on last line)
        self._keys.pop(self._key_index(key))



from xz import XZFile


def test_read_all(integration_case, data_pattern):
    xz_path, metadata = integration_case
    with XZFile(xz_path) as xz_file:
        streams_items = list(
            xz_file._fileobjs.items()  # pylint: disable=protected-access
        )
        assert len(streams_items) == len(metadata["streams"])
        pos = 0
        stream_boundaries = []
        block_boundaries = []
        for stream_item, metadata_stream in zip(streams_items, metadata["streams"]):
            stream_boundaries.append(pos)
            stream_pos, stream = stream_item
            assert stream_pos == pos
            assert stream.check == metadata_stream["check"]
            block_items = list(
                stream._fileobjs.items()  # pylint: disable=protected-access
            )
            assert len(block_items) == len(metadata_stream["blocks"])
            for block_item, metadata_block in zip(
                block_items, metadata_stream["blocks"]
            ):
                block_boundaries.append(pos)
                block_pos, block = block_item
                assert block_pos == pos - stream_pos
                assert (  # pylint: disable=protected-access
                    block._length == metadata_block["length"]
                )
                pos += metadata_block["length"]
            stream_length = pos - stream_pos
            assert stream._length == stream_length  # pylint: disable=protected-access
        assert xz_file.stream_boundaries == stream_boundaries
        assert xz_file.block_boundaries == block_boundaries
        assert xz_file.read() == data_pattern



import pytest

from xz.io import IOAbstract


def test_abstract_attributes():
    obj = IOAbstract(10)
    assert obj.readable() is True
    assert obj.seekable() is True
    assert obj.writable() is False


def test_abstract_tell_seek():
    obj = IOAbstract(10)
    assert obj.tell() == 0

    # absolute (no whence)
    assert obj.seek(1) == 1
    assert obj.tell() == 1
    assert obj.seek(3) == 3
    assert obj.tell() == 3
    assert obj.seek(10) == 10
    assert obj.tell() == 10
    with pytest.raises(ValueError) as exc_info:
        obj.seek(-1)
    assert str(exc_info.value) == "invalid seek position"
    with pytest.raises(ValueError) as exc_info:
        obj.seek(11)
    assert str(exc_info.value) == "invalid seek position"

    # absolute (with whence)
    assert obj.seek(5, 0) == 5
    assert obj.tell() == 5
    assert obj.seek(10, 0) == 10
    assert obj.tell() == 10

    # relative
    assert obj.seek(-7, 1) == 3
    assert obj.tell() == 3
    assert obj.seek(2, 1) == 5
    assert obj.tell() == 5
    with pytest.raises(ValueError) as exc_info:
        obj.seek(-6, 1)
    assert str(exc_info.value) == "invalid seek position"
    with pytest.raises(ValueError) as exc_info:
        obj.seek(6, 1)
    assert str(exc_info.value) == "invalid seek position"

    # from end
    assert obj.seek(0, 2) == 10
    assert obj.tell() == 10
    assert obj.seek(-4, 2) == 6
    assert obj.tell() == 6
    assert obj.seek(-10, 2) == 0
    assert obj.tell() == 0
    with pytest.raises(ValueError) as exc_info:
        obj.seek(1, 2)
    assert str(exc_info.value) == "invalid seek position"
    with pytest.raises(ValueError) as exc_info:
        obj.seek(-11, 2)
    assert str(exc_info.value) == "invalid seek position"

    # from error
    with pytest.raises(ValueError) as exc_info:
        obj.seek(42, 3)
    assert str(exc_info.value) == "unsupported whence value"


def test_abstract_tell_read():
    class Impl(IOAbstract):
        def __init__(self):
            super().__init__(10)

        def _read(self, size):
            # for tests, does not rely on position
            return b"xyz"[:size]

    obj = Impl()

    # read all
    assert obj.read() == b"xyzxyzxyzx"
    obj.seek(5)
    assert obj.read() == b"xyzxy"

    # read from pos
    obj.seek(3)
    assert obj.read(2) == b"xy"
    assert obj.read(2) == b"xy"
    assert obj.read(2) == b"xy"
    assert obj.read(2) == b"x"
    assert obj.read(2) == b""
    assert obj.read(2) == b""


def test_abstract_tell_read_empty():
    class Impl(IOAbstract):
        def __init__(self):
            super().__init__(10)
            self.empty_reads = 100

        def _read(self, size):
            self.empty_reads -= 1
            if self.empty_reads > 0:
                return b""
            return b"a"

    obj = Impl()

    assert obj.read() == b"aaaaaaaaaa"



from io import TextIOWrapper

from xz.file import XZFile


class _XZFileText(TextIOWrapper):
    def __init__(self, filename, mode, encoding, errors, newline):
        self.xz_file = XZFile(filename, mode.replace("t", ""))
        super().__init__(self.xz_file, encoding, errors, newline)

    @property
    def stream_boundaries(self):
        return self.xz_file.stream_boundaries

    @property
    def block_boundaries(self):
        return self.xz_file.block_boundaries


def xz_open(filename, mode="rb", *, encoding=None, errors=None, newline=None):
    """Open an XZ file in binary or text mode.

    filename can be either an actual file name (given as a str, bytes,
    or PathLike object), in which case the named file is opened, or it
    can be an existing file object to read from or write to.

    For binary mode, this function is equivalent to the XZFile
    constructor: XZFile(filename, mode, ...). In this case, the
    encoding, errors and newline arguments must not be provided.

    For text mode, an XZFile object is created, and wrapped in an
    io.TextIOWrapper instance with the specified encoding, error
    handling behavior, and line ending(s).
    """
    if "t" in mode:
        if "b" in mode:
            raise ValueError(f"Invalid mode: {mode}")

        return _XZFileText(filename, mode, encoding, errors, newline)

    if encoding is not None:
        raise ValueError("Argument 'encoding' not supported in binary mode")
    if errors is not None:
        raise ValueError("Argument 'errors' not supported in binary mode")
    if newline is not None:
        raise ValueError("Argument 'newline' not supported in binary mode")

    return XZFile(filename, mode)



import json
from pathlib import Path

import pytest


@pytest.fixture(
    params=(Path(__file__).parent / "files").rglob("*.json"),
    ids=lambda path: path.name,
)
def integration_case(request):
    json_path = request.param
    with json_path.open() as json_file:
        metadata = json.load(json_file)
    return (json_path.with_suffix(".xz"), metadata)



try:
    from xz._version import __version__
except ImportError:  # pragma: no cover
    __version__ = "0.0.0.dev0-unknown"

from xz.common import XZError
from xz.file import XZFile
from xz.open import xz_open

# pylint: disable=redefined-builtin
open = xz_open
# pylint: enable=redefined-builtin


__all__ = (
    "__version__",
    "XZError",
    "XZFile",
    "open",
)
